/* Tamito KAJIYAMA <18 September 2004> */

/*
 * Note:
 * - Display resolution: Skencil = 72 dpi, m17n library = 100 dpi
 * - FT_Pos: a 32-bit integer represting a 26.6 fixed point number
 */

#include "Python.h"

#include <m17n-gui.h>
#include <m17n-misc.h>
#include <freetype/freetype.h>
#include <freetype/ftglyph.h>

#include <string.h>

static MFrame *frame;
static MDrawGlyph *mglyphs;
static int mglyphs_size;
static MDrawControl control;

static PyObject *_convert_outline(FT_Outline);

static char skm17n_init__doc__[] = "";

static PyObject *
skm17n_init(PyObject *self, PyObject *args)
{
  MPlist *plist;
  MFace *face;

  if (!PyArg_ParseTuple(args, ""))
    return NULL;
  if (!frame) {
    M17N_INIT();
    if (merror_code != MERROR_NONE) {
      PyErr_Format(PyExc_RuntimeError, "cannot initialize m17n library");
      return NULL;
    }
    plist = mplist();
    mplist_put(plist, Mdevice, NULL);
    frame = mframe(plist);
    m17n_object_unref(plist);
    if (!frame) {
      PyErr_Format(PyExc_RuntimeError, "cannot create frame");
      return NULL;
    }
    face = mframe_get_prop(frame, Mface);
    mface_put_prop(face, Msize, (void *)(10240 * 72 / 100));
    mface_put_prop(face, Mfontset, mfontset("truetype"));
    memset(&control, 0, sizeof(control));
    control.two_dimensional = 1;
    control.enable_bidi = 1;
    /* control.ignore_formatting_char = 1; */
    control.max_line_width = 0;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

static char skm17n_text_path__doc__[] = "";

static PyObject *
skm17n_text_path(PyObject *self, PyObject *args)
{
  unsigned char *utf8, *lang, *family, *font_type;
  int utf8_len, nchars, n, last, i, j;
  MText *mt;
  MFace *mf;
  void *fontp;
  FT_Face face;
  FT_GlyphSlot slot;
  PyObject *glyphs;
  int x, y;

  if (!PyArg_ParseTuple(args, "s#ss", &utf8, &utf8_len, &lang, &family))
    return NULL;
  if (!frame) {
    PyErr_Format(PyExc_RuntimeError, "m17n library not initialized");
    return NULL;
  }
  
  /* convert a UTF-8 string into a list of MDrawGlyph instances */
  mt = mtext_from_data(utf8, utf8_len, MTEXT_FORMAT_UTF_8);
  nchars = mtext_len(mt);
  if (strlen(lang) > 0)
    mtext_put_prop(mt, 0, nchars, Mlanguage, msymbol(lang));
  if (strlen(family) > 0) {
    mf = mframe_get_prop(frame, Mface);
    mface_put_prop(mf, Mfamily, msymbol(family));
  }
  mdraw_glyph_list(frame, mt, 0, nchars, &control, NULL, 0, &n);
  mglyphs = malloc(sizeof(MDrawGlyph) * (n + 1));
  if (mdraw_glyph_list(frame, mt, 0, nchars, &control, mglyphs, n, &n) < 0) {
    m17n_object_unref(mt);
    PyErr_Format(PyExc_RuntimeError, "mdraw_glyph_list() failed");
    return NULL;
  }
  mglyphs_size = n;
  mglyphs[n].fontp = NULL; /* sentinel */
  m17n_object_unref(mt);

  glyphs = PyTuple_New(mglyphs_size);
  x = y = 0;
  fontp = mglyphs[0].fontp;
  last = 0;
  for (i = 1; i <= mglyphs_size; i++) {
    if (fontp != mglyphs[i].fontp) {
      if (mglyphs[last].fontp == NULL ||
          mglyphs[last].font_type != Mfreetype) {
        if (mglyphs[last].fontp == NULL)
          font_type = "null";
        else if (mglyphs[last].font_type == Mx)
          font_type = "X";
        else if (mglyphs[last].font_type == Mxft)
          font_type = "Xft";
        else
          font_type = "unknown";
        PyErr_Format(PyExc_ValueError, "unsupported font type: %s", font_type);
        return NULL;
      }
      face = (FT_Face)mglyphs[last].fontp;
      for (j = last; j < i; j++) {
        if (FT_Load_Glyph(face, mglyphs[j].glyph_code, FT_LOAD_NO_BITMAP)) {
          PyErr_Format(PyExc_RuntimeError, "cannot load glyph: %d",
                       mglyphs[j].glyph_code);
          return NULL;
        }
        slot = face->glyph;
        if (slot->format != ft_glyph_format_outline) {
          PyErr_Format(PyExc_ValueError, "unsupported glyph format: %c%c%c%c",
                       (slot->format >> 24) & 255, (slot->format >> 16) & 255,
                       (slot->format >> 8)  & 255, slot->format & 255);
          return NULL;
        }
        FT_Outline_Translate(&slot->outline,
                             (FT_Pos)((x + mglyphs[j].x_off) * 64),
                             (FT_Pos)((y - mglyphs[j].y_off) * 64));
        PyTuple_SetItem(glyphs, j, _convert_outline(slot->outline));
        x += mglyphs[j].x_advance;
        y += mglyphs[j].y_advance;
      }
      fontp = mglyphs[i].fontp;
      last = i;
    }
  }
  free(mglyphs);
  return glyphs;
}

static PyObject *
_convert_outline(FT_Outline ft_outline)
{
  PyObject *py_outline, *contour, *point;
  int i, j, k;

  py_outline = PyTuple_New(ft_outline.n_contours);
  for (i = j = 0; i < ft_outline.n_contours; i++) {
    contour = PyTuple_New(ft_outline.contours[i] - j + 1);
    for (k = 0; j <= ft_outline.contours[i]; k++, j++) {
      point = Py_BuildValue("ffi",
                            (double)ft_outline.points[j].x / 64,
                            (double)ft_outline.points[j].y / 64,
                            ft_outline.tags[j] & 1);
      PyTuple_SetItem(contour, k, point);
    }
    PyTuple_SetItem(py_outline, i, contour);
  }
  return py_outline; /* new reference */
}

#define meth(name, func, doc) {name, (PyCFunction)func, METH_VARARGS, doc}

static struct PyMethodDef skm17n_methods[] = {
  meth("init",      skm17n_init,      skm17n_init__doc__),
  meth("text_path", skm17n_text_path, skm17n_text_path__doc__),
  {NULL, (PyCFunction)NULL, 0, NULL} /* sentinel */
};

static char skm17n_module__doc__[] = "";

void
initskm17n(void)
{
  PyObject *m;

  m = Py_InitModule4("skm17n", skm17n_methods, skm17n_module__doc__,
                     (PyObject*)NULL,PYTHON_API_VERSION);
  if (PyErr_Occurred())
    Py_FatalError("cannot initialize skm17n module");
}
